/*
 * Copyright (c) 2005,2015 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <pthread.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <neobotix/socketcan.h>
#include <neobotix/neo.h>
#include <neobotix/canMotors.h>

#include "canIds.h"
#include "driveParam.h"
#include "harmonica.h"
#include "neoTime.h"
#include "utils.h"


#define NMTNodeID 0x00
#define HeartbeatTimeMS 1000
/**
 ** Interface with the Elmo 'harmonica' motor controller 
 **/
/*----------------------------------------------------------------------*/
/*
 * Send a command to the byte interpreter (PDO2)
 */
int
intprtSetInt(CAN_HARMONICA_STR *h, int dataLen, char cmdChar1, char cmdChar2,
    int index, int data)
{
	CAN_MSG m;
	
	m.id = h->canOpen.rxPDO2;
	m.len = dataLen;
	m.type = 0;
	m.data[0] = cmdChar1;
	m.data[1] = cmdChar2;

	m.data[2] = index & 0xff;
	m.data[3] = index >> 8;
	
	m.data[4] = data & 0x000000ff;
	m.data[5] = (data & 0x0000ff00) >> 8;
	m.data[6] = (data & 0x00ff0000) >> 16;
	m.data[7] = (data & 0xff000000) >> 24;
	/* printf("intprtSetInt: ");
	   canMsgShow(stdout, &m); */
	return socketcanTransmitMsg(h->dev, &m);
}

/*----------------------------------------------------------------------*/
/*
 * Service Data Object (SDO) transmission
 *
 * CANopen DS301 Implementation guide page 4-2 
 */
static void
sendSDODownload(CAN_HARMONICA_STR *h, int objIndex, int subIndex, int data)
{
	CAN_MSG m;
	const int ciInitDownloadReq = 0x20;
	const int ciNrBytesNoData = 0x00;
	const int ciExpedited = 0x02;
	const int ciDataSizeInd = 0x01;

	m.len = 8;
	m.type = 0;
	m.id = h->canOpen.rxSDO;
	m.data[0] = ciInitDownloadReq | (ciNrBytesNoData << 2) 
	    | ciExpedited | ciDataSizeInd;
	m.data[1] = (objIndex & 0x00ff);
	m.data[2] = (objIndex & 0xff00) >> 8;
	m.data[3] = subIndex;
	m.data[4] = (data & 0x000000ff);
	m.data[5] = (data & 0x0000ff00) >> 8;
	m.data[6] = (data & 0x00ff0000) >> 16;
	m.data[7] = (data & 0xff000000) >> 24;
	socketcanTransmitMsg(h->dev, &m);
}

/*----------------------------------------------------------------------*/
/*
 * Compute a velocity estimation, based on the position difference
 */
static double
estimVel(CAN_HARMONICA_STR *h, double pos)
{
	double vel;
	double dt;

	clock_gettime(CLOCK_REALTIME, h->currentTime);
	dt = absDiffTime(h->currentTime, h->velCalcTime);
	vel = (pos - h->oldPos)/dt;
	h->oldPos = pos;
	clock_gettime(CLOCK_REALTIME, h->velCalcTime);
	return vel;
}

/*----------------------------------------------------------------------*/
/*
 * Interpret the contents of the controller's status register
 */
static int
evalStatusRegister(CAN_HARMONICA_STR *h, int status)
{
	int ret;

	if (isBitSet(status, 0)) {
		if (h->motorState != ST_MOTOR_FAILURE) {
			fprintf(stderr, "Error 0x%x of drive %d: ",
				status, h->driveParam.driveIdent);
			if ((status & 0x0000000e) == 2) 
				fprintf(stderr, 
				    "- drive error under voltage\n");
			if ((status & 0x0000000e) == 4)
				fprintf(stderr, 
				    "- drive error over voltage\n");
			if ((status & 0x0000000e) == 10) 
				fprintf(stderr, 
				    "- drive error short circuit\n");
			if ((status & 0x0000000e) == 12)
				fprintf(stderr, 
				    "- drive error overheating\n");
			/* request detailed description of failure */
			intprtSetInt(h, 4, 'M', 'F', 0, 0);
		}
		h->newMotorState = ST_MOTOR_FAILURE;
		ret = -1;
	} else if (isBitSet(status, 6)) {
		/* Motor failure */
		if (h->motorState != ST_MOTOR_FAILURE) {
			fprintf(stderr, "Motor %d failure latched\n",
			    h->driveParam.driveIdent);
			/* request detailed description of failure */
			intprtSetInt(h, 4, 'M', 'F', 0, 0);
			clock_gettime(CLOCK_REALTIME, h->failureStartTime);
		}
		h->newMotorState = ST_MOTOR_FAILURE;
		ret = -1;
	} else {
		/* no error */
		ret = 0;
		/* general status bits */
		if (isBitSet(status, 4)) {
			if (h->motorState != ST_OPERATION_ENABLED) {
				fprintf(stderr, "Motor %d operation enabled\n",
				    h->driveParam.driveIdent);
				clock_gettime(CLOCK_REALTIME,
				    h->failureStartTime);
			}
			h->newMotorState = ST_OPERATION_ENABLED;
		} else {
			if (h->motorState != ST_OPERATION_DISABLED) {
				fprintf(stderr,
				    "Motor %d operation disabled\n",
				    h->driveParam.driveIdent);
			}
			h->newMotorState = ST_OPERATION_DISABLED;
		}
		/* current limit */
		if (isBitSet(status, 13)) {
			if (!h->currentLimitOn) {
				fprintf(stderr, "Motor %d current limit on\n",
				    h->driveParam.driveIdent);
			}
			h->currentLimitOn = true;
		} else {
			h->currentLimitOn = false;
		}
	}
	/* change state */
	h->motorState = h->newMotorState;
	return ret;
}

/*----------------------------------------------------------------------*/
/*
 * Interprets motor failure error codes
 */
static void
evalMotorFailure(CAN_HARMONICA_STR *h, int failure)
{
	fprintf(stderr, "Motor %d: ", h->driveParam.driveIdent);
	if (isBitSet(failure, 2)) {
		fprintf(stderr, "feedback loss\n");
	} else if (isBitSet(failure, 3)) {
		fprintf(stderr, "peak current exceeded\n");
	} else 	if (isBitSet(failure, 7)) {
		fprintf(stderr, "speed track error\n");
	} else 	if (isBitSet(failure, 8)) {
		fprintf(stderr, "position track error\n");
	} else 	if (isBitSet(failure, 17)) {
		fprintf(stderr, "speed limit exceeded\n");
	} else if (isBitSet(failure, 21)) {
		fprintf(stderr, "motor stuck");
	} else if (isBitSet(failure, 12)) {
		switch ((failure & 0xe000) >> 13) {
		case 0:
			fprintf(stderr, "OK\n");
			break;
		case 1:
			fprintf(stderr, "under voltage\n");
			break;
		case 2:
			fprintf(stderr, "over voltage\n");
			break;
		case 5:
			fprintf(stderr, "short circuit\n");
			break;
		case 6:
			fprintf(stderr, "overheating\n");
			break;
		default:
			fprintf(stderr, "reserved error code ??\n");
			break;
		}
	} else {
		fprintf(stderr, "unknown failure 0x%x\n", failure);
	}
}

/*----------------------------------------------------------------------*/
/*
 * Request status from the motor contoller 
 */
void 
harmonicaRequestStatus(CAN_HARMONICA_STR *h)
{
	intprtSetInt(h, 4, 'S', 'R', 0, 0);
}

/*----------------------------------------------------------------------*/
/*
 * reset the motor controller
 */
int
harmonicaReset(CAN_HARMONICA_STR *h)
{
	/* Init motor */
	harmonicaInitCtrl(h);
	/* Start motor */
	if (harmonicaStart(h) == -1) {
		return -1;
	}
	return 0;
}

/*----------------------------------------------------------------------*/
/* 
 * shut down the motor controller
 */
void
harmonicaStop(CAN_HARMONICA_STR *h)
{
	intprtSetInt(h, 8, 'M', 'O', 0, 0);
}

/*----------------------------------------------------------------------*/
/*
 * Initialize communication with the motor controller
 */
CAN_HARMONICA_STR *
harmonicaInit(int dev, CAN_MOTOR_ID id)
{
	CAN_HARMONICA_STR *h;
	
	h = (CAN_HARMONICA_STR *)malloc(sizeof(CAN_HARMONICA_STR));
	if (h == NULL) {
		return NULL;
	}
	h->dev = dev;
	h->divForRequestStatus = 10;
	h->canTimeout = 6;
	h->statusCtrl = 0;
	h->posGearMeasRad = 0;
	h->angleGearRadMem = 0;
	h->velGearMeasRadS = 0;
	h->motionType = MOTIONTYPE_VELCTRL;

	h->started = (pthread_cond_t *)malloc(sizeof(pthread_cond_t));
	pthread_cond_init(h->started, NULL);
	
	h->currentTime = (struct timespec *)malloc(sizeof(struct timespec));
	h->watchdogTime = (struct timespec *)malloc(sizeof(struct timespec));
	h->velCalcTime = (struct timespec *)malloc(sizeof(struct timespec));
	h->failureStartTime = 
	    (struct timespec *)malloc(sizeof(struct timespec));
	h->sendTime = (struct timespec *)malloc(sizeof(struct timespec));

	clock_gettime(CLOCK_REALTIME, h->velCalcTime);

	h->limSwLeft = false;
	h->limSwRight = false;
	
	h->countRequestDiv = 0;
	
	h->motorState = ST_PRE_INITIALIZED;
	h->currentLimitOn = false;

	h->numAttempsRecFail = 0;
	
	driveParamInit(&h->driveParam, id);
	
	switch (id) {
	case CAN_MOTOR_LEFT:
		h->canOpen.txPDO1 = TxPDO1_LEFT;
		h->canOpen.txPDO2 = TxPDO2_LEFT;
		h->canOpen.rxPDO2 = RxPDO2_LEFT;
		h->canOpen.txSDO = TxSDO_LEFT;
		h->canOpen.rxSDO = RxSDO_LEFT;
		break;
	case CAN_MOTOR_RIGHT:
		h->canOpen.txPDO1 = TxPDO1_RIGHT;
		h->canOpen.txPDO2 = TxPDO2_RIGHT;
		h->canOpen.rxPDO2 = RxPDO2_RIGHT;
		h->canOpen.txSDO = TxSDO_RIGHT;
		h->canOpen.rxSDO = RxSDO_RIGHT;
		break;
	} /* switch */

	clock_gettime(CLOCK_REALTIME, h->sendTime);
	clock_gettime(CLOCK_REALTIME, h->currentTime);

	return h;
}

/*----------------------------------------------------------------------*/
/*
 * Ends the commnunication with a motor controller 
 */
void
harmonicaEnd(CAN_HARMONICA_STR *h)
{
	harmonicaStop(h);

	free(h->started);
	free(h->currentTime);
	free(h->watchdogTime);
	free(h->velCalcTime);
	free(h->failureStartTime);
	free(h->sendTime);

	free(h);
}

/*----------------------------------------------------------------------*/
/*
 * request initialization of the motor controller
 */
void
harmonicaInitCtrl(CAN_HARMONICA_STR *h)
{
	CAN_MSG msg;

	/* (re)start  CAN bus */
	canMsgSet(&msg, 0, 1, 0, 0, 0, 0, 0, 0, 0);
	msg.len = 2;
	socketcanTransmitMsg(h->dev, &msg);
	usleep(10000);
	h->motorState = ST_PRE_INITIALIZED;

	/* Set position counter to zero */
	intprtSetInt(h, 8, 'P', 'X', 0, 0);
	usleep(10000);
	/*
	 * Set PDO mapping : 
	 * TPD01: 
	 *  - position
	 *  - velocity
	 */
	/* stop all emissions of TPD01 */
	sendSDODownload(h, 0x1a00, 0, 0);
	
	/* position 4 byte of TPD01 */
	sendSDODownload(h, 0x1a00, 1, 0x60640020);
	
	/* velocity 4 byte of TPD01 */
	sendSDODownload(h, 0x1a00, 2, 0x60690020);

	/* transmission type "synch" */
	sendSDODownload(h, 0x1800, 2, 1);

	/* activate mapped objects */
	sendSDODownload(h, 0x1a00, 0, 2);

	harmonicaRequestStatus(h);
}

/*----------------------------------------------------------------------*/
/*
 * start the motor controller
 */
int
harmonicaStart(CAN_HARMONICA_STR *h)
{
	int cnt;
	struct timespec ts;

	/* start motor */
	intprtSetInt(h, 8, 'M', 'O', 0, 1);	
	harmonicaRequestStatus(h);
	
	/* wait a most 10s for motor to become enabled */
	cnt = 0;
	while (h->motorState != ST_OPERATION_ENABLED && cnt < 10) {
		cnt++;
		if (h->parent != NULL) {
			clock_gettime(CLOCK_REALTIME, &ts);
			ts.tv_sec++;
			pthread_cond_timedwait(h->started,
			    h->parent->mutex, &ts);
		} else
			usleep(500000);
		intprtSetInt(h, 8, 'M', 'O', 0, 1);	
		harmonicaRequestStatus(h);
		
	} /* while */
	if (h->motorState != ST_OPERATION_ENABLED) {
		return -1;
	}
	clock_gettime(CLOCK_REALTIME, h->watchdogTime);
	clock_gettime(CLOCK_REALTIME, h->sendTime);
	return 0;
}
/*----------------------------------------------------------------------*/
/*
 * Activate the hardware watchdog
 */
void
harmonicaStartWatchdog(CAN_HARMONICA_STR *h, bool started)
{

	if (started) {
		/* consumer (PC) heartbeat time */
		sendSDODownload(h, 0x1016, 1, 
		    (NMTNodeID << 16) | HeartbeatTimeMS);
		/* error behaviour after failure */
		sendSDODownload(h, 0x1029, 1, 2);
		/* motor behaviour after heartbeat failure: "quick stop" */
		sendSDODownload(h, 0x6007, 0, 3);
		/* activate emergency events: "heartbeat event" */
		sendSDODownload(h, 0x2f21, 0, 0x8);
	} else {
		/* Stop watchdog */
	}
}

/*----------------------------------------------------------------------*/
/*
 * Request a given position and velocity and ask for data updates
 */
void 
harmonicaSetGearPosVelRadS(CAN_HARMONICA_STR *h, 
    double posGearRad, double velGearRadS)
{
	int posEncIncr;
	int velEncIncrPeriod;
	CAN_DRIVE_PARAMS *p = &h->driveParam;
	CAN_MSG msg;

	driveParamPosVelRadToIncr(p, posGearRad, velGearRadS,
	    &posEncIncr, &velEncIncrPeriod);
	posEncIncr *= driveParamGetSign(p);
	velEncIncrPeriod *= driveParamGetSign(p);
	if (velEncIncrPeriod > driveParamGetVelMax(p)) {
		velEncIncrPeriod = driveParamGetVelMax(p);
	}
	if (velEncIncrPeriod < -driveParamGetVelMax(p)) {
		velEncIncrPeriod = -driveParamGetVelMax(p);
	}
	h->refVelGearIncS = velEncIncrPeriod;
	if (h->motionType == MOTIONTYPE_POSCTRL) {
		if (velEncIncrPeriod != 0) {
			intprtSetInt(h, 8, 'J', 'V', 0, velEncIncrPeriod);
			intprtSetInt(h, 4, 'B', 'G', 0, 0);
		} else {
			intprtSetInt(h, 8, 'P', 'A', 0, posEncIncr);
			intprtSetInt(h, 4, 'B', 'G', 0, 0);
		}
	} else if (h->motionType == MOTIONTYPE_VELCTRL) {
		intprtSetInt(h, 8, 'J', 'V', 0, velEncIncrPeriod);
		intprtSetInt(h, 4, 'B', 'G', 0, 0);
	}
	/* request pos and vel by TPDO1, triggerd by SYNC msg */
	canMsgSet(&msg, 0x80, 0, 0, 0, 0, 0, 0, 0, 0);
	msg.len = 0;
	socketcanTransmitMsg(h->dev, &msg);
}

/*----------------------------------------------------------------------*/
/*
 * request a given velocity, and ask for data updates
 * also ask for status every divForRequestStatus calls
 */
void
harmonicaSetGearVelRadS(CAN_HARMONICA_STR *h, double vel)
{
	int velEncIncrPeriod;
	CAN_DRIVE_PARAMS *p = &h->driveParam;
	CAN_MSG m;
	double dt;

	velEncIncrPeriod = p->sign 
	    * driveParamVelGearRadSToVelMotIncrPeriod(p, vel);

	if (velEncIncrPeriod > driveParamGetVelMax(p)) 
		velEncIncrPeriod = driveParamGetVelMax(p);
	if (velEncIncrPeriod < -driveParamGetVelMax(p))
		velEncIncrPeriod = -driveParamGetVelMax(p);
	h->refVelGearIncS = velEncIncrPeriod;

	intprtSetInt(h, 8, 'J', 'V', 0, velEncIncrPeriod);
	intprtSetInt(h, 4, 'B', 'G', 0, 0);
	
	/* request pos and vel by TPDO1, triggered by SYNC msg */
	m.id = 0x80;
	m.len = 0;
	m.type = 0;
	memset(m.data, 0, 8);
	socketcanTransmitMsg(h->dev, &m);
	
	/* send heartbeat */
	m.id = 0x700;
	m.len = 5;
	m.type = 0;
	memset(m.data, 0, 8);
	socketcanTransmitMsg(h->dev, &m);

	clock_gettime(CLOCK_REALTIME, h->currentTime);
	dt = absDiffTime(h->currentTime, h->sendTime);
	if (dt > 1.1) {
		fprintf(stderr, 
		    "Time between send velocity of motor %d is too large %f\n",
		    driveParamGetDriveIdent(p), dt);
	}
	clock_gettime(CLOCK_REALTIME, h->sendTime);
	/* request status */
	h->countRequestDiv++;
	if (h->countRequestDiv > h->divForRequestStatus) {
		harmonicaRequestStatus(h);
		h->countRequestDiv = 0;
	}
}

/*----------------------------------------------------------------------*/
/*
 * Display status of a motor controller on stdout
 */
void
harmonicaShow(CAN_HARMONICA_STR *h)
{
	printf("motionType: %d\n", h->motionType);
	printf("statusCtrl: 0x%08x\n", h->statusCtrl);
	printf("ref velocity: %d (%f)\n", h->refVelGearIncS,
	       driveParamVelMotIncrPeriodToVelGearRadS(&h->driveParam, 
						       h->refVelGearIncS));
	printf("velocity (rad/S): %f\n", h->velGearMeasRadS);
	printf("position (rad): %f\n", h->posGearMeasRad);
	switch (h->motorState) {
	case ST_PRE_INITIALIZED:
	        printf("state: PRE_INITIALIZED\n");
		break;
	case ST_OPERATION_ENABLED:
	        printf("state: OPERATION_ENABLED\n");
		break;
	case ST_OPERATION_DISABLED:
	        printf("state: OPERATION_DISABLED\n");
		break;
	case ST_MOTOR_FAILURE:
	        printf("state: MOTOR_FAILURE\n");
		break;
	default:
	        printf("state: %d (unknown)\n", h->motorState);
	}
	printf("current limit: %s\n", h->currentLimitOn ? "On" : "Off");
};

/*----------------------------------------------------------------------*/

/*
 * Decode CAN messages sent by the motor controller
 *
 * XXX there are 3 different ways of evaluating position and velocity ???
 */
#define MaxNumAttempsRecFail 1
#define WaitTimeAfterFail 5.0

bool
harmonicaDecode(CAN_HARMONICA_STR *h, CAN_MSG *msg)
{
	bool ret =  false;
	int temp1, temp2, digIn, posCnt, failure;
	CAN_DRIVE_PARAMS *p = &h->driveParam;

	if (msg->id == h->canOpen.txPDO1) {
		/* Position and velocity update */
		temp1 = canMsgGetInt(msg, 0);
		h->posGearMeasRad = driveParamGetSign(p) 
		    * driveParamPosMotIncrToPosGearRad(p, temp1);
		temp2 = canMsgGetInt(msg, 4);
		h->velGearMeasRadS = driveParamGetSign(p) 
		    * driveParamVelMotIncrPeriodToVelGearRadS(p, temp2);
		clock_gettime(CLOCK_REALTIME, h->watchdogTime);
		ret = true;
	} 
	if (msg->id == h->canOpen.txPDO2) {
		if (msg->data[0] == 'P' && msg->data[1] == 'X') {
			/* current pos */
			posCnt = canMsgGetInt(msg, 4);
			h->posGearMeasRad = driveParamGetSign(p)
			    * driveParamPosMotIncrToPosGearRad(p, posCnt);
			h->velGearMeasRadS = estimVel(h, h->posGearMeasRad);
		} else if (msg->data[0] == 'P' && msg->data[1] == 'A') {
			/* position absolute */
		} else if (msg->data[0] == 'J' && msg->data[1] == 'V') {
			/* ?? */
		} else if (msg->data[0] == 'B' && msg->data[1] == 'G') {
			/* begin motion */
		} else if (msg->data[0] == 'U' && msg->data[1] == 'M') {
			/* user mode */
			digIn = 0x1fffff & canMsgGetInt(msg, 4);
			if ((digIn & 0x0400) != 0x0000) {
				h->limSwRight = true;
			}
		} else if (msg->data[0] == 'S' && msg->data[1] == 'R') {
			/* status */
			h->statusCtrl = canMsgGetInt(msg, 4);
			ret = evalStatusRegister(h, h->statusCtrl);
			if (ret == 0) {
				/* signal that motor is OK */
				pthread_cond_signal(h->started);
			}
		} else if (msg->data[0] == 'M' && msg->data[1] == 'F') {
			/* motor failure */
			failure = canMsgGetInt(msg, 4);
			evalMotorFailure(h, failure);
		} else if (msg->data[0] == 'M' && msg->data[1] == 'O') {
			/* Motor On reply */
			canMsgGetInt(msg, 4);
		} else if (msg->data[0] == 'U' && msg->data[1] == 'M') {
			/* debug eval */
			canMsgGetInt(msg, 4);
		} else if (msg->data[0] == 'P' && msg->data[1] == 'M') {
			canMsgGetInt(msg, 4);
		} else if (msg->data[0] == 'A' && msg->data[1] == 'C') {
			canMsgGetInt(msg, 4);
		} else if (msg->data[0] == 'D' && msg->data[1] == 'C') {
			canMsgGetInt(msg, 4);
		} else {
			fprintf(stderr, "harmonicaDecode: unknown msg from PDO02 %c%c\n",
			    msg->data[0], msg->data[1]);
		}
		clock_gettime(CLOCK_REALTIME, h->watchdogTime);
		ret = true;
	}
	if (msg->id == h->canOpen.txSDO) {
		posCnt = driveParamGetSign(p)*canMsgGetInt(msg, 4);
		h->posGearMeasRad = 
		    driveParamPosMotIncrToPosGearRad(p, posCnt);
		h->velGearMeasRadS = estimVel(h, h->posGearMeasRad);
		clock_gettime(CLOCK_REALTIME, h->watchdogTime);
		ret = true;
	}
	/* try to recover from failure */
	if (h->motorState == ST_MOTOR_FAILURE) {
		if (h->numAttempsRecFail < MaxNumAttempsRecFail) {
			clock_gettime(CLOCK_REALTIME, h->currentTime);
			if (absDiffTime(h->currentTime, h->failureStartTime)
			    > WaitTimeAfterFail) {
				/* reset motor */
				ret = harmonicaReset(h);
				clock_gettime(CLOCK_REALTIME, 
				    h->failureStartTime);
				if (ret == 0) {
					h->numAttempsRecFail = 0;
				} else {
					h->numAttempsRecFail++;
				}
			}
		}
	}
				
	return ret;
}
